#ifndef lint
#ifdef SUNOS4
static char rcsid[] =
"@(#) $Header: main.c,v 1.9 92/11/02 21:03:44 leres Exp $ (LBL with SUNOS4)";
#else
static char rcsid[] =
"@(#) $Header: main.c,v 1.9 92/11/02 21:03:44 leres Exp $ (LBL without SUNOS4)";
#endif
#endif
/*
 * tconsole
 * Copyright 1990 Tom Lawrence, Brown University
 * Last Modification 6/14/90
 */
#include <sys/types.h>

#include <ctype.h>
#include <errno.h>
#include <signal.h>
#include <stdio.h>

#include "global.h"
#include "ascii.h"
#include "get_char.h"

extern tty_pair init_tty();
extern char *ttyname();
tty_pair main_tty;
int errno;

/*****************************************************************************
handle input from the keyboard. Keystrokes are simply passed on to the
tconsole pty.
*****************************************************************************/
handle_kbd_input(pty)
	int pty;
{
	char buf[128];
	int size;

	if ((size = read(0, buf, sizeof(buf))) < 1) {
		/* 
		 * This should be done with blocking i/o, but for some reason
		 * it doesn't work under SunOS 4.1 unless in non-blocking mode
		 */
		if (errno != EWOULDBLOCK) {
			perror("tconsole: handle_kbd_input: read");
			exit(1);
		} else
			return;
	}
	if (write(pty, buf, size) != size) {
		perror("tconsole: write");
		exit(1);
	}
}

/*****************************************************************************
handle output from the tconsole pty. get_char() is called initially in
a mode which fills its buffer but returns nothing useful. It is then
called in a loop in such a manner that it returns data until its
buffer is empty; it then returns an error code and this routine exits.
Printable chars are buffered up to the right edge of the screen and
then sent to print_string(). If an ESC is encountered in the output,
the ESC sequence handler is called.  Other non printable chars are
handled separately.
*****************************************************************************/
handle_pty_output(pty)
	int pty;
{
	int c, length;
	char buf[256];

	/* erase the cursor */

	putcursor();

	/* fill the input buffer, but return nothing (except maybe an error
	 * code) */
	if (get_char(pty, READ | NBIO) < 0)
		return;

	do {
		length = 0;

		do {
			if ((c = get_char(pty, RET_DATA)) != -1) {

				/* buffer printables up to the right edge of
				 * the screen */

				if (isprint(c) && xtab + length < window_width_cols)
					buf[length++] = c;
			}
		}

		/* until no more chars to read, a non-printable was reached, or
		 * we've run out of space on this line */

		while (c != -1 && isprint(c) && length < (sizeof(buf) - 1) &&
		    xtab + length < window_width_cols);

		/* print what we've got so far */

		buf[length] = 0;
		print_string(buf, length);

		/* take care of non-printables */

		if (c == ESC)
			handle_esc(pty);

		else if (!isprint(c))
			handle_ctrl_char(c, pty);

	} while (c != -1);

	putcursor();
}

/*****************************************************************************
This routine is called when the child shell exits. We clear the utmp
entry and exit.
*****************************************************************************/
byebye()
{
	FILE *f;
	char *tty = "/dev/tty";

	putcursor();		/* erase cursor */
	setup_utmp(main_tty.tty, 1);

	/* Open tty to clean things up a bit */
	if ((f = fopen(tty, "w")) == 0) {
		fprintf(stderr, "tconsole: byebye fopen");
		perror(tty);
		exit(1);
	}

#ifdef notdef
	/* XXX can't do this because there's no way to erase the prom cursor */
	/* Cursor prom emulator to where were were in the graphics plane */
	fprintf(f, "\033[%d;%dH", ytab + 1, xtab + 1);
#else
	/* Erase stuff left in graphics plane... */
	fprintf(f, "\033[0J");
#endif
	fflush(f);

	exit(0);
}

/*****************************************************************************
*****************************************************************************/
main(argc, argv)
	int argc;
	char **argv;
{
	fd_set fdset;

	/* parse options */
	options(argc, argv);

	/* bad things happen if this is run on top of X or sunview */
	if (!debug && strcmp("/dev/console", ttyname(0))) {
		fputs("tconsole can only be run on the system console.\n",
		    stderr);
		exit(1);
	}

	/* set up the window, pty/tty, screen and utmp entry */
	create_win();
	main_tty = init_tty();
	init();
	setup_utmp(main_tty.tty, 0);

	/* watch for the child shell exiting */
	signal(SIGCHLD, byebye);
	signal(SIGINT, byebye);
	signal(SIGHUP, byebye);

	/*
	 * Arrange for keyboard strokes to go to /dev/kbd instead of
	 * /dev/console and attach stdin to /dev/kbd
	 */
	kbd_mode();

	putcursor();

	/* manage data flow */
	while (1) {
		FD_ZERO(&fdset);
		FD_SET(main_tty.pty, &fdset);
		FD_SET(0, &fdset);

		if (select(getdtablesize(), &fdset, 0, 0, 0) < 1) {
			perror("tconsole: select");
			exit(1);
		}
		if (FD_ISSET(main_tty.pty, &fdset))
			handle_pty_output(main_tty.pty);
		if (FD_ISSET(0, &fdset))
			handle_kbd_input(main_tty.pty);
	}
}
